# * ====================================================================
# * Copyright (c) 1995,1996, ISIS/JRC.  All rights reserved.
# *
# * This product was developed within the Institute for Safety, Informatics
# * and Systems of the Joint Research Centre of the European Commission.
# *
# * More information about us; and this software product can
# * be found on http://www.jrc.it.
# * 
# * With this file you should have received a file called 'License'.
# * 
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions
# * are met:
# *
# * 1. Redistributions of source code must retain the above copyright
# *    notice, this list of conditions and the following disclaimer. 
# *
# * 2. Redistributions in binary form must reproduce the above copyright
# *    notice, this list of conditions and the following disclaimer in
# *    the documentation and/or other materials provided with the
# *    distribution.
# *
# * 3. All materials mentioning features or use of this
# *    software must display the following acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * 4. The name of the Copyright holder must not be used to
# *    endorse or promote products derived from this software without
# *    prior written permission.
# *
# * 5. Redistributions of any form whatsoever must retain the following
# *    acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * THIS SOFTWARE IS PROVIDED BY ISIS/JRC ``AS IS'' AND ANY
# * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ISIS/JRC OR
# * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# * OF THE POSSIBILITY OF SUCH DAMAGE.
# * ====================================================================
#

package ServerConfig;

use strict;

use ConfigFile;
use Z3950;
use Z3950::Z3950;


# Verification of arguments and conversion to
# standard names
#
%ServerConfig::ServerQuirks = (
   'nonamedresultsets' => 'nonamedresultsets',
   'nofullelementset'  => 'nofullelementset',
);
%ServerConfig::DbaseQuirks = (
   'keyvaluetree'  => 'keyvaluetree',
   'coloncentered' => 'coloncentered',
   'dif'           => 'dif',

   'dashtospace'   => 'dashtospace',
   'correctcase'   => 'correctcase',
   'htmlcontent'   => 'htmlcontent'
);


# Support functions to check the input
#


sub ServerConfig::PortNumber { 
   my ($key, $port) = @_;
   
   if ( $port =~ m/^\d+$/ ) {
      return ($key, $port);
   } elsif ( lc($port) eq 'z3950' or lc($port) eq 'z39.50' ) {
      return ($key, $Z3950::DEFAULT_PORT);
   } else {
      return (undef, "invalid port $port, specify a port number or 'z39.50'");
   }
}
sub ServerConfig::ServerType {
   my ($key, $type) = @_;

   return (undef, undef)
      unless $type;

   if ( ! eval "use ASN1::$type;" ) {
      return (undef, "Invalid type '$type', $@");
   }
   return ($key, $type);
}
sub ServerConfig::AttributeSets {
   my ($key, $value) = @_;

   # convert a list of 'geo', 'bib1', etc.
   # into \$geo_Use, \$Bib1_Use, etc

   return &ConfigFile::VerifyArray($key, $value, \%Z3950::AttributeSets);
}
sub ServerConfig::RecordSyntax {
   my ($key, $value) = @_;

   # convert a list of 'grs1', 'sutrs', etc.
   # into \$OID_RS_GRS1, \$OID_RS_SUTRS, etc

   return &ConfigFile::VerifyArray($key, $value, \%Z3950::RecordSyntaxes);
}
sub ServerConfig::ServerQuirks {
   my ($key, $value) = @_;

   # convert a list of 'dashtospace', 'htmlcontent', etc.
   # into the same, but verify whether they are allowed.

   return &ConfigFile::VerifyArray($key, $value, \%ServerConfig::ServerQuirks);
}
sub ServerConfig::DbaseQuirks {
   my ($key, $value) = @_;

   # convert a list of 'dashtospace', 'htmlcontent', etc.
   # into the same, but verify whether they are allowed.

   return &ConfigFile::VerifyArray($key, $value, \%ServerConfig::DbaseQuirks);
}
sub ServerConfig::DbaseName {
   my ($key, $dbase) = @_;
   
   if ( $dbase =~ m/^[a-zA-Z0-9_-]+$/ ) {
      return ($key, $dbase);
   } else {
      return(undef, "Invalid dbase '$dbase'");
   }
}

# Structure of the input files
#
%ServerConfig::Dbase = (
   'title'        => [],
   'description'  => [],
   'categories'   => [\&ConfigFile::CommaArray],
   'quirks'       => [\&ConfigFile::CommaArray, \&ConfigFile::Lowercase,
                      \&ServerConfig::DbaseQuirks,
                      \&ConfigFile::ArrayToHash],
   'ID'           => 'Dbase'
);

%ServerConfig::Server = (
   'host'         => [],
   'port'         => [\&PortNumber],
   'username'     => [],
   'password'     => [],
   'type'         => [\&ServerType],
   'title'        => [],
   'description'  => [],
   'attributes'   => [\&ConfigFile::CommaArray, \&AttributeSets],
   'categories'   => [\&ConfigFile::CommaArray],
   'recordsyntax' => [\&ConfigFile::CommaArray, \&RecordSyntax],
   'dbase'        => [\&ServerConfig::DbaseName,
                      \%ServerConfig::Dbase],
   'quirks'       => [\&ConfigFile::CommaArray, \&ConfigFile::Lowercase,
                      \&ServerConfig::ServerQuirks,
                      \&ConfigFile::ArrayToHash],
   'ID'           => 'Server',
   'DEFAULT'      => {'port'     => 210,
                      'type'     => 'BER'
                     }
);

1; # End of package ServerConfig

