#!/usr/local/bin/perl
# * ====================================================================
# * Copyright (c) 1995,1996, ISIS/JRC.  All rights reserved.
# *
# * This product was developed within the Institute for Safety, Informatics
# * and Systems of the Joint Research Centre of the European Commission.
# *
# * More information about us; and this software product can
# * be found on http://www.jrc.it.
# * 
# * With this file you should have received a file called 'License'.
# * 
# * Redistribution and use in source and binary forms, with or without
# * modification, are permitted provided that the following conditions
# * are met:
# *
# * 1. Redistributions of source code must retain the above copyright
# *    notice, this list of conditions and the following disclaimer. 
# *
# * 2. Redistributions in binary form must reproduce the above copyright
# *    notice, this list of conditions and the following disclaimer in
# *    the documentation and/or other materials provided with the
# *    distribution.
# *
# * 3. All materials mentioning features or use of this
# *    software must display the following acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * 4. The name of the Copyright holder must not be used to
# *    endorse or promote products derived from this software without
# *    prior written permission.
# *
# * 5. Redistributions of any form whatsoever must retain the following
# *    acknowledgment:
# *    "This product includes software developed by ISIS/JRC"
# *
# * THIS SOFTWARE IS PROVIDED BY ISIS/JRC ``AS IS'' AND ANY
# * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL ISIS/JRC OR
# * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# * OF THE POSSIBILITY OF SUCH DAMAGE.
# * ====================================================================
#
# PerlZ, Z client in Perl
#


# Paranoid we are
#
use strict;

# Initial setup and default settings
#
$| = 1;

# hostname lookups seem to be non trivial
use Sys::Hostname;
use Socket;
use IO::Select;
use FileHandle;
use Cwd;

use ASN1;

use Z3950::Z3950;
use Z3950;

use Z3950::ASN1::RecordSyntax_SUTRS;
use Z3950::ASN1::RecordSyntax_generic;
use Z3950::ASN1::RecordSyntax_opac;
use Z3950::ASN1::RecordSyntax_summary;

use Display;
use PerlZ;
use ServerConfig;
use ConfigFile;
use POSIX "sys_wait_h";


# Global variables
#
$::IDSTRING = "PerlZ (id=116)";
$::NAMESTRING = "PerlZ, STA/ISIS";
$::VERSIONSTRING = "0.3";
$::progname = ($0 =~ m/([\w\.]+)$/)?$1:$0;

%::connections = ();
$::current_connection_id = 0;
$::divider = '';

$::force = 0;

my $arg;
while ( $arg = shift @ARGV ) {
   if ( $arg eq '-c' and $PerlZ::CONFIGFILENAME = shift @ARGV ) {
      # nop
   } elsif ( $arg eq '-D' ) {
      my $debuglist = shift @ARGV;
      $ASN1::debug = $::debug = 1
         if $debuglist =~ s/\bmain\b//ig or !$debuglist;
      $ASN1::debug_received = 1
         if $debuglist =~ s/\breceived\b//ig;
      $ASN1::debug_sent = 1
         if $debuglist =~ s/\bsent\b//ig;
      $ASN1::BER::debug = 1
         if $debuglist =~ s/\bBER\b//ig;
      $Z3950::debug = 1
         if $debuglist =~ s/\bZ\b//ig;
      $Display::debug = $Display::HTML::debug = $Display::HGS::debug = 1
         if $debuglist =~ s/\bdisplay\b//ig;
      $ConfigFile::debug = 1
         if $debuglist =~ s/\bconfig\b//ig;

      $debuglist =~ s/^\s*//;
      $debuglist =~ s/\s*$//;
      $debuglist =~ s/\s*,\s*//g;
      warn "Ignored debugging flags: '$debuglist'"
         if $debuglist;
   } elsif ( $arg eq '-f' ) {
      $::force = 1;
   } else {
      warn "Syntax error: $arg ".join(' ', @ARGV) . "\n"
         unless $arg eq '-h';
      die "Syntax: $0 [-D <debugging>] [-c <configfile>]\n"
        . "    -D    Switch on debuggin: main,received,sent,BER,Z,display,config,stacking\n"
        . "    -f    force overwriting of the socket\n"
        . "    -c    Change the name of the config file [$PerlZ::CONFIGFILENAME]\n";
   }
}



# Read configuration file
#
$::config = &ConfigFile::Read($PerlZ::CONFIGFILENAME,
                              \%PerlZ::ConfigFile::Main)
   or die "Could not read config file $PerlZ::CONFIGFILENAME";


# Copy the appropriate values into the modules
#
$ASN1::CONNECTION_TIMEOUT = $::config->{timeout};
$Z3950::CONNECTION_LIFETIME = $::config->{lifetime};
$Z3950::SERVERSDIR = $::config->{serversdir};


# Signal handling
#
$SIG{QUIT} = \&::leave;
$SIG{INT}  = \&::leave;
$SIG{TERM} = \&::leave;
$SIG{HUP}  = 'IGNORE';
$SIG{PIPE} = 'IGNORE';


# Try to bind to port
#
$::serversocket = new FileHandle;

if ( -e $::config->{socketname} ) {
   if ( !-S $::config->{socketname} or !$::force) {
      warn "$::config->{socketname} already exists";
      warn "$::config->{socketname} is not a socket"
         if !-S $::config->{socketname};
      $::config->{socketname} = undef;       # avoid the file being unlinked
      exit 1;
   } else {
      unlink $::config->{socketname};
   }
}

socket($::serversocket, PF_UNIX, SOCK_STREAM, 0)
   or die "socket failed, $!";

my $retry = 0;
my $maxtries = 100;
while ( !bind($::serversocket, sockaddr_un($::config->{socketname}))
        and $retry < $maxtries ) {
   warn "bind to port $::config->{port} failed, $!, retrying (try $retry out of $maxtries)";
   $retry++;

   sleep 1;
}

die "bind failed"
   if $retry == $maxtries;

warn "-- Listening on socket $::config->{socketname}"
   if $::debug;

chmod(0777, $::config->{socketname})
   or warn "chmod (0777, $::config->{socketname}) failed, $!";

listen($::serversocket, 10)
   or die "Listen failed";
setsockopt($::serversocket, SOL_SOCKET, &SO_REUSEADDR, 1)
   or warn "Could not set setsockopt(SO_REUSEADDR): $!";


# Perform the conversion of template files to the final ones
# (changing port numbers, host names, warnings, etc)
#
my $filename;
foreach $filename ( @{ $::config->{templatefiles} } ) {
   &ProcessTemplate($filename);
}

# my $prevFD = select $::serversocket; $| = 1;
# select $prevFD;

warn "$::NAMESTRING version $::VERSIONSTRING restarted";

# Now handle all the incoming connections
# and write the results back as HTTP
#
while (1) {
   while ( waitpid(-1, POSIX::WNOHANG) != 0 and waitpid(-1, POSIX::WNOHANG) != -1 ) {};    # Clean up zombies

   my $rin = '';
   map { vec($rin, fileno($_), 1) = 1 if $_ }
        ($::serversocket, map { $::connections{$_}->{parentsocket} } keys %::connections);
   my $nfound = select($rin, undef, undef, 10);

   next           # timeout
      if $nfound == 0;

   &leave('Select failed '.$!)
      if $nfound == -1;

   if ( $nfound and vec($rin, fileno($::serversocket), 1) ) {
      # Incoming client connection

      my $clientsocket = new FileHandle;
      my $client = accept($clientsocket, $::serversocket)
         or die "Could not accept connection, $!";

      select $clientsocket; $| = 1;

      my $errormsg = &HandleConnection($clientsocket);

      if ($errormsg) {
         syswrite($clientsocket, $errormsg, length($errormsg));
         close $clientsocket;
      }
   }


   # Check all the sockets running from the children for available
   # data
   #
   foreach my $conn_id ( keys %::connections ) {
      last
         if $nfound == 0;

      next
         if !vec($rin, fileno($::connections{$conn_id}->{parentsocket}), 1);

      $nfound--;

      my $buffer;
      my $in = sysread($::connections{$conn_id}->{parentsocket},
                       $buffer, 10000);

      if ( $in == -1 ) {
         warn "Error $! while reading from child $conn_id";
         &CloseChildConnection($::connections{$conn_id}->{parentsocket});
      } elsif ( $in == 0 ) {
         warn "-- Child $conn_id closed connection"
            if $::debug;
         &CloseChildConnection($::connections{$conn_id}->{parentsocket});
      } elsif ( !$::connections{$conn_id}->{clientsocket} ) {
         warn "Child $conn_id is babbling, discarding $in bytes";
         &CloseChildConnection($::connections{$conn_id}->{parentsocket});
      } else {
         if ( $buffer =~ s/\s$::divider$//os ) {      # EOT encountered?
            syswrite($::connections{$conn_id}->{clientsocket},
                     $buffer, length($buffer));
            close $::connections{$conn_id}->{clientsocket};
            delete $::connections{$conn_id}->{clientsocket}
         } else {
            syswrite($::connections{$conn_id}->{clientsocket},
                     $buffer, length($buffer));
         }
      }
   }
}

sub CloseChildConnection {
   my ($parentsocket) = @_;

   my ($conn_id);

   # Zap all the entries that are managed through this parent-child connection
   #
   foreach $conn_id ( keys %::connections ) {
      if ( $::connections{$conn_id}->{parentsocket} eq $parentsocket ) {
         close $::connections{$conn_id}->{clientsocket}
            if $::connections{$conn_id}->{clientsocket};
         delete $::connections{$conn_id};
      }
   }
}

sub HandleConnection {
   my ($clientsocket) = @_;

   my ($errormsg, $header, $arguments) = &ProcessHeader($clientsocket);

   warn "Connection from $header->{REMOTE_ADDR}";

   # Set defaults
   #
   $arguments->{protocol} = 'HTML'
      unless $arguments->{protocol};

   warn "-- Server: $arguments->{server}, DBase: $arguments->{dbase},"
      . " Protocol: $arguments->{protocol}, ConnId: $arguments->{conn_id}"
      if $::debug;

   my $conn_id;         # the index into %::connections

   if ( $errormsg ) {
      warn "Connection from $header->{REMOTE_ADDR} closed";
      return &Display::HTML::ErrorPage(500, $errormsg);
   } else {
      # Select the child to be used or create a new one
      if ( $conn_id = $arguments->{conn_id} and $::connections{$conn_id} ) {
         $::connections{$conn_id}->{clientsocket} = $clientsocket;
         warn "-- Using existing connection for $conn_id";
      } else {
         $conn_id = &CreateChild($clientsocket);
         if (!$conn_id) {
            warn "Connection from $header->{REMOTE_ADDR} closed";
            return &Display::HTML::ErrorPage(500, 'Server too busy');
         }
         $arguments->{conn_id} = $conn_id;
      }

      # transmit request
      my $parentsocket = $::connections{$conn_id}->{parentsocket};
      my $arg;
      foreach $arg ( keys %{ $arguments } ) {
         next
            if $arg =~ m/^\s*$/;
         my $message = "$arg: $arguments->{$arg}\n";
         syswrite($parentsocket, $message, length($message));
      }
      syswrite($parentsocket, "\n", 1);    # empty line (separator)
      return undef;
   }
}


# ProcessHeader
#  Processes the incoming CGIsock request and, in case of success,
#  returns the headers and split arguments. Otherwise an error is returned.
#
sub ProcessHeader {
   my ($socket) = @_;

   # 4 things pass over the socket
   # web server -> script: Environment variables as 'Key: Value' lines
   # script -> web server: 'ACCEPTED <optional message>'
   # web server -> script: Content of client message (POST and PUT
   #                       methods for example)
   # script -> web server: HTTP HEADER (Content-type, etc.) + HTML page


   # Read the environment variables from the socket
   # and parse them into a hash.
   #
   my $environment = &socket_read($socket);

   my %header = undef;   # clear the environment;
   foreach my $line ( split /[\r\n]+/s, $environment) {
          my ($key, $value) = split /=/, $line, 2;
          $header{$key} = $value;
   }

   # Note the kruft after ACCEPTED is for testing purposes only...  The
   # only key word the CGISOCK wants is ACCEPTED, though the command must
   # be the first word in the response. This is so CGISOCK tolerant of
   # trailing rubbish without having to resort to regex to grok your
   # commands

   my $message = "ACCEPTED by $::NAMESTRING v$::VERSIONSTRING\n";
   syswrite($socket, $message, length($message));

   # Read the message posted by the web client to the web server
   # The script _must_ read the message from the client, even if
   # client did not send a message at all. If there is no message
   # mog_cgisock will send the string 'CGISOCK: no client block'.
   #
   my $content = &socket_read($socket);

   # Parse what we read into something useful

   if ( $header{REQUEST_METHOD} eq 'POST' ) {
      ($content) = split m/\n/, $content, 1;    # only use the first line
      $header{QUERY_STRING} = $content;
   } elsif ( $header{REQUEST_METHOD} eq 'GET' ) {
      # nop
   } else {
      return "Unknown method '$header{REQUEST_METHOD}'";
   }

   warn "-- URL is now '$header{SCRIPT_URI}?$header{QUERY_STRING}'"
      if $::debug;

   # Split the arguments into key:value pairs
   #
   my %arguments = &ProcessArguments($header{QUERY_STRING});

   # Create a lower case entry of each argument
   #
   my $key;
   foreach $key (keys %arguments) {
      $arguments{lc($key)} = $arguments{$key};
   }

   return (undef, \%header, \%arguments);         # no errormsg and the header and arg hashes
}


# ProcessArguments
#  Process the passed HTTP request into key=value pairs.
#
sub ProcessArguments {
   my ($arguments) = @_;
   my %out = undef;
   my $arg;

   return undef
      if !$arguments;

   foreach $arg ( split('&', $arguments) ) { 
      # Rely on the browser to sensibly escape the equal sign
      # and the ampersand, although not all browsers (Microsoft)
      # do this correctly. But if this is wrong, there is no
      # way for this script to correct, or even detect
      # that error :-(
      #
      my ($nam, $val) = split m/=/, $arg, 2;
      $val = 1
         if !defined($val);

      $nam =~ s/\+/ /g ;
      $val =~ s/\+/ /g ;
      $val =~ s/%(..)/pack('c',hex($1))/eg;
      $nam =~ s/%(..)/pack('c',hex($1))/eg; 

      # Remove any leading/postfixed white space for
      # the variable name... MS-IE 3.0 has a tendency
      # to add this.
      $nam =~ s/^\s+//g;
      $nam =~ s/\s+$//g;

      # Remove leading and trailing whitespace from value
      $val =~ s/^\s+//g;
      $val =~ s/\s+$//g;

      warn "Key occurs again: \$out{$nam} = '$out{$nam}' overwritten with '$val'"
         if defined($out{$nam});
      $out{$nam} = $val;
   }

   return %out;
}

sub socket_read ($;$) {
        my ($socket, $maxlen) = @_;

        die "Undefined socket passed, $maxlen"
                unless $socket;

        my @blocks = ();
        my $block;
        my $lentotal = 0;

        while ( new IO::Select($socket)->can_read ) {
                my $len = 4096;
                $len = $maxlen - $lentotal
                        if defined($maxlen) and $len+$lentotal > $maxlen;

                my $lenread = sysread($socket, $block, $len);
                last
                        unless $lenread;
                push @blocks, $block;
                $lentotal += $lenread;
                last
                        unless $lenread == $len;
                warn "Read $lenread bytes"
                        if $::debug;
        }
        warn "-- Read $lentotal bytes (total)"
                if $::debug;
        return join('', @blocks);
}


# CreateChild
#  Create a child and hand off the connection to it.
#  Closes all the unnecessary files before continuing.
#  The child never leaves this function, the parent does.
#  The function returns the connection Id or undef if no
#  connection is built up.
#
sub CreateChild {
   my ($clientsocket) = @_;

   my $parentsocket = new FileHandle;
   my $childsocket = new FileHandle;
   if (!socketpair($parentsocket,$childsocket,PF_UNIX,SOCK_STREAM,PF_UNSPEC)) {
         warn "Could not create socketpair, $!";
         return undef;
   }

   my $pid = fork;
   if ( $pid ) {
      # parent here
      close $childsocket;
      my $prevFD = select $parentsocket; $|=1; select $prevFD;

      fcntl $parentsocket, &F_SETFL, &O_NONBLOCK;
      if ( time() > $::current_connection_id ) {
         $::current_connection_id = time();
      } else {
         $::current_connection_id++;
      }
      $::connections{$::current_connection_id} = {
            parentsocket=>$parentsocket,
            clientsocket=>$clientsocket,
            pid=>$pid,
            conn_id=>$::current_connection_id
      };
      return $::current_connection_id;
   } elsif ( defined($pid) ) {
      # child here
      close $parentsocket;    # close all file descriptors of the parent
      close $clientsocket;    # in the child except childsocket
      close $::serversocket;
      $::serversocket = undef;
      map { close $::connections{$_}->{parentsocket};
            close $::connections{$_}->{clientsocket}
               if $::connections{$_}->{clientsocket};
            delete $::connections{$_}
          } keys %::connections;

      my $prevFD = select $childsocket; $|=1; select $prevFD;
      &HandleRequests($childsocket);

      warn "-- Clean up the kitchen and leave"
         if $::debug;
      &Z3950::CloseConnections();
      exit 0;
   } else {
      warn "Can't fork, $!";
      # should return error code to the recipient
      close $parentsocket;
      close $childsocket;
      close $clientsocket;
      return undef;
   }
}



# Two stages in leaving, first ask politely, on second kill kill violently
#
sub leave {
   my ($signal) = @_;

   $SIG{QUIT} = \&::bailout;     # make sure we catch a second INTR, etc.
   $SIG{INT}  = \&::bailout;
   $SIG{TERM} = \&::bailout;

   # pretty pretty pretty-print
   #
   $signal = "Signal $signal received"
      if defined($SIG{$signal});
   $signal .= '. '
      if $signal;

   warn $signal.'Leaving...';

   &Z3950::CloseConnections();

   # ask all the children to get rid of themselves
   my @pids = grep { $_ } map { $::connections{$_}->{pid} } keys %::connections;
   warn 'Terminating '.join(', ', @pids)
      if $::debug and $#pids != -1;
   kill 'TERM', @pids
      if $#pids != -1;

   unlink($::config->{socketname})
      if $::serversocket and -S $::config->{socketname};

   exit(1);
}

sub bailout {
   my ($signal) = @_;

   # pretty pretty pretty-print
   #
   $signal = "Signal $signal received"
      if defined($SIG{$signal});
   $signal .= '. '
      if $signal;
   warn $signal.'Bailing out...';

   # nuke all the children
   my @pids = grep { $_ } map { $::connections{$_}->{pid} } keys %::connections;
   warn 'Nuking '.join(", ", @pids)
      if $::debug and $#pids != -1;
   kill 'KILL', @pids
      if $#pids != -1;

   unlink($::config->{socketname})
      if $::serversocket and -S $::config->{socketname};

   exit(1);
}


###############################################################################
# Child #######################################################################
###############################################################################

sub HandleRequests {
   my ($socket) = @_;
   my $errormsg = '';
   my ($rmsg, $line, $rin);

   do {
      # Read a request from the input socket and all its arguments
      #
      $rin = '';
      vec($rin, fileno($socket), 1) = 1;
      my $nfound = select($rin, undef, undef, 300);
      if ( $nfound == -1 ) {           # serious error, let's bail out
         warn 'Select failed, '.$!;
         last;
      }

      next                             # select timed out -> no arguments
         if $nfound == 0;

      my $arguments = undef;
      while( $line = <$socket> ) {
         last
            if $line =~ m/^\s*$/;
         my ($key, $value) = split m/\s*:\s*/, $line, 2;
         chomp $value;
         $arguments->{lc($key)} = $value;
      }
      
      last                 # socket was closed (EOF -> empty input)
         if !$arguments;

      warn "-- Request for connection $arguments->{conn_id} received"
         if $::debug;

      if ( $arguments->{protocol} eq 'HGS' ) {
         $rmsg = &HandleHGSSearch($arguments);
      } elsif ( $arguments->{protocol} eq 'HTML' ) {
         if ( $arguments->{firstresult} ) {
            $rmsg = &HandleHTMLFetch($arguments);
         } else {
            $rmsg = &HandleHTMLSearch($arguments);
         }
      }
      syswrite($socket, $rmsg.$::divider, length($rmsg.$::divider));

      &Z3950::TimeStampConnection($::connections{$arguments->{conn_id}}->{handle});
   } while &Z3950::CheckConnections();
}

sub HandleHGSSearch {
   my ($arguments) = @_;

   my $conn_id     = $arguments->{conn_id}
      or return &Display::HGS::ErrorPage(500, 'No connection Id');
   my $server      = $arguments->{server}
      or return &Display::HGS::ErrorPage(200, 'No server specified');
   my $dbase       = $arguments->{dbase}
      or return &Display::HGS::ErrorPage($server, 'No dbase specified for '.$server);

   my $serverconfig = &GetServerConfig($server)
      or return &Display::HGS::ErrorPage($server, "No server found by name of $server");

   # Create the query from the input parameters
   #
   if ( $::debug ) {
      print STDERR "-- Query:";
      print STDERR " Text='$arguments->{text}'"
         if defined($arguments->{text});
      print STDERR " LatMin='$arguments->{latmin}'"
         if defined($arguments->{latmin});
      print STDERR " LatMax='$arguments->{latmax}'"
         if defined($arguments->{latmax});
      print STDERR " LonMin='$arguments->{lonmin}'"
         if defined($arguments->{lonmin});
      print STDERR " LonMax='$arguments->{lonmax}'"
         if defined($arguments->{lonmax});
      warn " ";
   }

   my ($field, $relation, @query, @applied) = ();

   if ( $arguments->{text}
        and $field = &Z3950::Field('Any', @{ $serverconfig->{attributes} }) ) {
      my @text = ();
      # Take the quoted ones.
      while ($arguments->{text} =~ s/(["'])(.*?)\1// ) { push @text, $2 };
      # and the single words
      my @text;
      push @text, grep {$_} split m/\s+/, $arguments->{text};
      # and concatentate the terms into one big 'or'
      # FIXME Should this be 'and'. This way we'll have more results, so leave it for now
      @query = &Z3950::RPNOr(map { &Z3950::RPNOperand($_, $field) } @text);
      push @applied, 'text';
   }

   if ( $arguments->{lonmin}
        and $field = &Z3950::Field('West Bounding Coordinate', @{ $serverconfig->{attributes} })
        and $relation = &Z3950::Relation('greater or equal') ) {
      push @query, &Z3950::RPNOperand($arguments->{lonmin}, $field, $relation);
      push @applied, 'lonmin';
   }
   if ( $arguments->{lonmax}
        and $field = &Z3950::Field('East Bounding Coordinate', @{ $serverconfig->{attributes} })
        and $relation = &Z3950::Relation('less than or equal') ) {
      push @query, &Z3950::RPNOperand($arguments->{lonmax}, $field, $relation);
      push @applied, 'lonmax';
   }
   if ( $arguments->{latmin}
         and $field = &Z3950::Field('South Bounding Coordinate', @{ $serverconfig->{attributes} })
         and $relation = &Z3950::Relation('greater or equal') ) {
      push @query, &Z3950::RPNOperand($arguments->{latmin}, $field, $relation);
      push @applied, 'latmin';
   }
   if ( $arguments->{latmax}
         and $field = &Z3950::Field('North Bounding Coordinate', @{ $serverconfig->{attributes} })
         and $relation = &Z3950::Relation('less than or equal') ) {
      push @query, &Z3950::RPNOperand($arguments->{latmax}, $field, $relation);
      push @applied, 'latmax';
   }

   my $query = &Z3950::RPNCreate($Z3950::OID_AS_Bib1, &Z3950::RPNAnd(@query))
      or return &Display::HGS::ErrorPage($server, 'No valid query specified');

   # Initialise the connection to the server
   #
   my $handle = &Z3950::InitConnection($serverconfig)
      or return &Display::HGS::ErrorPage($server, 'Could not connect to server: '.$@);
   $::connections{$conn_id}->{handle} = $handle;

   # Do the actual query and pick up results if not returned at the first try
   #
   my ($nr, @Records) = &PerformSearch($handle, $dbase, $query, -1);
   return &Display::HGS::ErrorPage($server, $Records[0], 'No results matching your query')
      if $nr == 0;

   if ( $nr > $#Records+1 ) {
      my ($nr2, @Records2) = &FetchResults($handle, $#Records+2, $nr-$#Records-1, 0);
      return &Display::HGS::ErrorPage($server, $Records[0], 'No results matching your query')
         if $nr2 == 0;
      push @Records, @Records2;
   }

   # Decode, display and return the result of the query
   #
   my $quirks = $serverconfig->{dbase}->{$dbase}->{quirks};
   my $i = 0;
   @Records = &DecodeRecords($handle, $quirks, @Records);
   @Records = map { &Display::HGS::Record($_,
                       &CreateURL($arguments->{conn_id}, $server, $dbase, ++$i, 1, 1))
                  } @Records;
   @Records = grep { $_ }  @Records;      # Filter out the empty ones

   return &Display::HGS::Page($server, $#Records+1, join("\n", @Records),
                              (@applied? ('Applied'=>join(' ', @applied)):undef));
}


sub HandleHTMLSearch {
   my ($arguments) = @_;

   # Check the input parameters
   #
   my $conn_id = $arguments->{conn_id}
      or return &Display::HTML::ErrorPage(500, 'No connection Id');
   my $server = $arguments->{server}
      or return &Display::HTML::ErrorPage(200, 'No server specified');
   my $dbase = $arguments->{dbase}
      or return &Display::HTML::ErrorPage(200, 'No database specified for '.$server);

   # Fetch the Id for the description of the server
   #
   my $serverconfig = &GetServerConfig($server)
      or return &Display::HTML::ErrorPage(200, 'No configuration info for '.$server);

   my $key;
   if ( $::debug ) {
      my $line = "-- Query:";
      foreach $key ( keys %{ $arguments } ) {
         next
            unless $key =~ m/^(text|latmin|latmax|lonmin|lonmax|key\d+|value\d+|op\d+)$/;
         $line .= " $key='$arguments->{$key}'"
            if $arguments->{$key};
         if ( length($line) > 80 ) {
            print STDERR $line;
            $line = "\n--";
         }
      }
      warn $line
         if length($line) > 3;
   }


   my ($field, $relation, @query) = ();

   if ( defined($arguments->{text}) and $field = &Z3950::Field('Any', @{ $serverconfig->{attributes} }) ) {
      my @text = ();
      # Take the quoted ones.
      while ($arguments->{text} =~ s/(["'])(.*?)\1// ) { push @text, $2 };
      # and the single words
      $arguments->{text} =~ s/^\s+//;     # cut off leading and
      $arguments->{text} =~ s/\s+$//;     # ... trailing whitespace
      # and convert them into separate terms
      @query = map { &Z3950::RPNOperand($_, $field) }
                   (@text, split m/\s+/, $arguments->{text});
      # and concatentate the terms into one big 'or'
      # FIXME Should this be 'and'? This way we'll have more results, so leave it for now
      @query = (&Z3950::RPNOr(@query));
   }

   # Process the fielded parameters
   #
   foreach $key ( keys %{ $arguments } ) {
      next
         unless $key =~ m/^value(\d+)$/;
      my $digit = $1;

      if ( $arguments->{"value$digit"} ) {
         $arguments->{"key$digit"} = 'any'
            unless $arguments->{"key$digit"};
         $relation = &Z3950::Relation($arguments->{"op$digit"});
         push @query, &Z3950::RPNOperand($arguments->{"value$digit"}, $field, $relation)
            if  $field = &Z3950::Field($arguments->{"key$digit"}, @{ $serverconfig->{attributes} });
      }
   }

   push @query, &Z3950::RPNOperand($arguments->{lonmin}, $field, $relation)
      if defined($arguments->{lonmin})
      and $field = &Z3950::Field('west bounding coordinate', @{ $serverconfig->{attributes} })
      and $relation = &Z3950::Relation('greater or equal');
   push @query, &Z3950::RPNOperand($arguments->{lonmax}, $field, $relation)
      if defined($arguments->{lonmax})
      and $field = &Z3950::Field('east bounding coordinate', @{ $serverconfig->{attributes} })
      and $relation = &Z3950::Relation('less than or equal');
   push @query, &Z3950::RPNOperand($arguments->{latmin}, $field, $relation)
      if defined($arguments->{latmin})
      and $field = &Z3950::Field('south bounding coordinate', @{ $serverconfig->{attributes} })
      and $relation = &Z3950::Relation('greater or equal');
   push @query, &Z3950::RPNOperand($arguments->{latmax}, $field, $relation)
      if defined($arguments->{latmax})
      and $field = &Z3950::Field('north bounding coordinate', @{ $serverconfig->{attributes} })
      and $relation = &Z3950::Relation('less than or equal');

   my $query = &Z3950::RPNCreate($Z3950::OID_AS_Bib1, &Z3950::RPNAnd(@query))
      or return &Display::HTML::ErrorPage(200, 'No valid query specified');
      
   my $handle = &Z3950::InitConnection($serverconfig)
      or return &Display::HTML::ErrorPage(500, 'Could not connect to server: '.$@);
   $::connections{$conn_id}->{handle} = $handle;

   my ($nr, @Records) = &PerformSearch($handle, $dbase, $query, 100);
   return &Display::HTML::ErrorPage(200,
                                 $Records[0], 'No results matching your query')
      if $nr == 0;

   if ( $nr > $#Records+1 ) {
      my ($nr2, @Records2) = &FetchResults($handle, $#Records+2, $nr-$#Records-1, 0);
      return &Display::HTML::ErrorPage(200, 'No results matching your query')
         if $nr2 == 0;
      push @Records, @Records2;
   }

   my $quirks = $serverconfig->{dbase}->{$dbase}->{quirks};
   @Records = &DecodeRecords($handle, $quirks, @Records);

   my $resulttxt = ($nr) . ' result' . ($nr == 1? '':'s');
   my $next = '&nbsp;';
   if ( $nr > @Records ) {
      my $n = $nr - @Records;
      $n = $::config->{maxresults}
         if $n > $::config->{maxresults};
      $next = '<A HREF="'.&CreateURL($conn_id,$server,$dbase,$#Records+1,$n,0).'">next'.($n == 1? '':" $n").'</A>';
   }

   my $i = 1;
   return &Display::HTML::Page(200, 
          $nr.' result'.($nr == 0? '':'s'),
          "<TABLE ALIGN=CENTER BORDER=0 WIDTH=100%><TR><TD ALIGN=RIGHT>$next</TD></TR></TABLE>"
          . '<TABLE CELLSPACING=0 CELLPADDING=0 WIDTH=90%>' . "\n"
          . join("\n", 
                map { '<TR><TD><B>Record&nbsp;'
                      . '<A HREF="'
                      . &CreateURL($conn_id,$server,$dbase,$i,1,1)
                      . '">'.$i++.'</A>'
                      . '</B></TD>'
                      . '<TD><HR SIZE=4 NOSHADE></TD></TR>'."\n"
                      . &Display::HTML::Record($_)
                    } @Records
               )
         . "\n</TABLE>"
         );
}


sub HandleHTMLFetch {
   my ($arguments) = @_;

   my $conn_id     = $arguments->{conn_id}
      or return &Display::HTML::ErrorPage(500, 'No connection Id');
   my $server      = $arguments->{server}
      or return &Display::HTML::ErrorPage(200, 'No server specified');
   my $dbase       = $arguments->{dbase}
      or return &Display::HTML::ErrorPage(200, 'No dbase specified for '.$server);

   my ($firstresult, $nresults, $full) = (1,1,0);
   $firstresult = $arguments->{firstresult}
      if $arguments->{firstresult};
   $nresults    = $arguments->{nresults}
      if $arguments->{nresults};
   $full        = $arguments->{full};

   warn "-- Fetching results $firstresult+$nresults" . ($full? ' as full records ':' ') . "from $server/$dbase"
      if $::debug;

   return &Display::HTML::ErrorPage(408, 'Connection has timed out, please repost query')
      unless exists $::connections{$conn_id}; 
   my $serverconfig = &GetServerConfig($server)
      or return &Display::HTML::ErrorPage(200, 'No configuration info for '.$server);
   my $handle = $::connections{$conn_id}->{handle}
      or return &Display::HTML::ErrorPage(500, 'No handle in connections hash: ');

   my ($nr, @Records) = &FetchResults($handle, $firstresult, $nresults, $full);
   return &Display::HTML::ErrorPage(200, $Records[0])
      if $nr == 0;

   my $i = $firstresult;

   my $quirks = $serverconfig->{dbase}->{$dbase}->{quirks};
   @Records = &DecodeRecords($handle, $quirks, @Records);
   my ($prev, $next) = ('&nbsp;', '&nbsp;');
   if ( $firstresult > 1 ) {
      my $n = $firstresult - $nresults;
      $n = 1
         if $n < 1;
      $prev = '<A HREF="'.&CreateURL($conn_id,$server,$dbase,$n,$nresults,$full).'">prev'.($nresults == 1? '':" $nresults").'</A>';
   }
   $next = '<A HREF="'.&CreateURL($conn_id,$server,$dbase,$firstresult+$nresults,$nresults,$full).'">next'.($nresults == 1? '':" $nresults").'</A>';


   return &Display::HTML::Page(200, 
             ($#Records+1).' result'.($#Records == 0? '':'s'),
             "<TABLE ALIGN=CENTER BORDER=0 WIDTH='100%'><TR><TD ALIGN=LEFT>$prev</TD><TD ALIGN=RIGHT>$next</TD></TR></TABLE>"
             . '<TABLE CELLSPACING=0 CELLPADDING=0 WIDTH=90%>'."\n"
             . join("\n", 
                  map { '<TR><TD><B>Record&nbsp;'
                      . '<A HREF="'
                      . &CreateURL($arguments->{conn_id},$server,$dbase,$i,1,1)
                      . '">'.$i++.'</A>'
                      . '</B></TD><TD><HR SIZE=4 NOSHADE></TD>'."\n"
                     . &Display::HTML::Record($_)
                  } @Records
               ) . "\n</TABLE>"
          );
}


# PerformSearch($hanlde, $dbase, $query, $nresults)
#  Preforms a search, does error trapping.
#  Returns ($n, @R) in case there are $n results and @R results returned
#  Returns (0, $errormsg) in case there 0 results or an error
#
sub PerformSearch {
   my ($handle, $dbase, $query, $nresults) = @_;

   # Verify and correct arguments
   #
   warn "-- Requested $nresults results, limitting to $::config->{maxresults}"
      if $::debug and $nresults > $::config->{maxresults};
   $nresults = $::config->{maxresults}
      if $nresults > $::config->{maxresults} or $nresults == -1;

   # Perform search
   #
   my $SR = &Z3950::Search($handle, $dbase, $query, 0, $nresults);

   # Do error trapping for search
   #
   if ( !$SR ) {
      return (0, $@);
   } elsif ( lc($SR->{searchStatus}) ne $BER::TRUE ) {
      warn "multipleNonSurrogateDiag found! Not that we used it or anything, no..."
         if $::debug and $SR->{multipleNonSurrogateDiag};

      my $errormsg = &Z3950::diagnostic($SR->{records}->{nonSurrogateDiagnostic},
                                               'Search failed (no diagnostic)');
      warn $errormsg
         if $::debug;
      return (0, $errormsg);
   }

   $SR->{records}->{responseRecords} = []        # prevent an undef value
      if !$SR->{records}->{responseRecords};

   if ( $SR->{numberOfRecordsReturned} != $#{$SR->{records}->{responseRecords}}+1 ) {
      warn "-- \$#results = $#{$SR->{records}->{responseRecords}}+1, but \$#recordsReturned = $SR->{numberOfRecordsReturned}, "
         if $Z3950::debug;
      $SR->{numberOfRecordsReturned} = $#{$SR->{records}->{responseRecords}}+1;
   }

   warn "-- $SR->{resultCount} results, $SR->{numberOfRecordsReturned} returned"
      if $::debug;

   return ($SR->{resultCount},
           ($SR->{numberOfRecordsReturned} > 0? @{ $SR->{records}->{responseRecords} }:()));
}


# FetchResults
#     Picks up the results from an existing resultset
#     $handle        handle of the connection
#     $firstresult   index of first result to pick up, [1..]
#
sub FetchResults {
   my ($handle, $firstresult, $nresults, $full) = @_;

   # Verify and correct arguments
   #
   $firstresult = 1        # correct the input value in case it is bogus (should not happen)
      if !$firstresult;

   return (0, ())
      if $nresults <= 0;

   warn "First result should be [1..] but is 0"
      if $::debug and $firstresult == 0;

   if ( $nresults > $::config->{maxresults} ) {
      warn "-- Requesting $nresults results, limitting to $::config->{maxresults}"
         if $::debug;
      $nresults = $::config->{maxresults};
   }

   # Iterate and pick up all the (remaining) results.
   # Try to pick them up all in one go. If the pickup fails, try to pick up 
   # fifty percent of the amount. Repeat this division by ten until you succeed
   # in getting the records or you end up retrieving 0 records at a time.
   #
   my @Records = ();
   my $n = 0;
   my $nr = $nresults;
   my $errormsg = undef;

   while ( $n < $nresults and $nr > 0 ) {
      my $PR = &Z3950::Present($handle, $full, $firstresult + $n, $nr);

      if ( !$PR ) {
         return (0, $@);
      } elsif ( $PR->{presentStatus} eq 'failure' ) {
         $errormsg = &Z3950::diagnostic($PR->{records}->{nonSurrogateDiagnostic},
                                        'Present failed (no diagnostic)');
         warn "-- $errormsg when fetching $nr results at the same time"
            . (int($nr/2) > 0? ", retrying with " . int($nr/2) : "" )
            if $::debug;

         $nr = int ($nr/2);        # try again, but with a lower number of records
      } elsif ( $PR->{numberOfRecordsReturned} == 0
                or ref $PR->{records}->{responseRecords} ne 'ARRAY' ) {
         if ( $#Records == -1 ) {
            return (0, 'Present did not return any result');
         } else {
            last;
         }
      } else {
         push @Records, @{ $PR->{records}->{responseRecords} };
         $n = $#Records+1;
         $nr = $nresults-$n
            if $nr > $nresults-$n;

         warn "-- " . ($#{ $PR->{records}->{responseRecords} }+1) . " results returned, ($n/$nresults)" . ($nr?", next fetch: $nr":'')
            if $::debug
      }
   }

   warn "Could only retrieve $n out of $nresults records"
      if $::debug and $n != $nresults;

   if ( $n == 0 ) {
      if ( $errormsg ) {
         push @Records, $errormsg;
      } else {
         push @Records, 'Not able to retreive any records';
      }
   }

   return ($n, @Records);
}


# DecodeRecords
#  Walks through a list of returned records and decodes them into a hash tree
#
sub DecodeRecords {
   my ($handle, $quirks, @Records) = @_;

   my $i;
   for ( $i = 0; $i <= $#Records; $i++ ) {
      if ( $Records[$i]->{surrogateDiagnostic} ) {   # o oh...
         my $errormsg = 'Present record returned error,<BR>'
                   . &Z3950::diagnostic($Records[$i]->{surrogateDiagnostic});
         warn $errormsg
            if $::debug;
         $Records[$i] = {'error'=>$errormsg};
      } elsif ( $Records[$i]->{record}->{retrievalRecord} ) {
         $Records[$i] = &Z3950::parseRecord($Records[$i]->{record}->{retrievalRecord}, $quirks, 'record '.($i+1));
      } else {
         my $errormsg = 'Cannot decode what has been returned';
         if ($::debug and eval 'use Data::Dumper;' ) {
            local $Data::Dumper::Indent = 1;
            warn &Dumper($Records[$i]);
         }
         warn $errormsg
            if $::debug;
         $Records[$i] = {'error'=>$errormsg};
      } 
   }

   return (@Records);
}


sub CreateURL {
   my ($conn_id, $server, $dbase, $firstresult, $nresults, $full) = @_;

   return 'http://'.$::config->{hostname}.':'.$::config->{port}
        . $::config->{urlroot} . '?'
        . join('&', 'protocol=HTML',
                    'dbase='.$dbase,
                    'server='.$server,
                    'conn_id='.$conn_id,
                    'firstresult='.$firstresult,
                    'nresults='.$nresults)
        . ($full? '&full=1':'');
}



# ProcessTemplate
#  Replaces the following strings with the appropriate texts
#  ###PERLZPORT###     current port number of server
#  ###PERLZHOST###     current host running the software
#  ###WARNING###         a message indicating the fact that the file is automagically created

sub ProcessTemplate {
   my ($dest) = @_;
   my ($startcomment, $continuecomment, $endcomment);

   my $source = $dest.'.tmpl';

   if ( $dest =~ m/\.html?$/ ) {          # HTML documents
      ($startcomment, $continuecomment, $endcomment) = ('<!--', '', '-->');
   } elsif ($dest =~ m/\.txt$/ ) {        # HGS directory
      ($startcomment, $continuecomment, $endcomment) = ('#', '#', '#');
   } else {                               # Wild guess
      warn "Unknown extension for '$dest', guessing comment delimiters"
         if $::debug;
      ($startcomment, $continuecomment, $endcomment) = ('#', '#', '#');
   }

   my $warning = $startcomment . "This file is automatically created by zc.pl.\n" . $continuecomment . "Edit $filename.tmpl instead.\n" . "$continuecomment\n"x39 . $endcomment;

   if ( !open TMPL, $source ) {
      warn "Could not open $source, $!";
      return;
   }
   if ( !open DEST, ">$dest" ) {
      warn "Could not open $dest for writing, $!";
      close TMPL;
      return;
   }

   my $line;
   while ( $line = <TMPL> ) {
      $line =~ s/###WARNING###/$warning/;
      $line =~ s/###PERLZHOST###/$::config->{hostname}/;
      $line =~ s/###PERLZPORT###/$::config->{port}/;
      print DEST $line;
   }
   close TMPL;
   close DEST;

   warn '-- Wrote new ' . $dest;
   return 1;
}



# Server configuration management
#


sub GetServerConfig {
   my ($server) = @_;

   if ( $server =~ m/([a-z0-9._-]+):(\d+|z39.?50)/ ) {
      return {'host'=>$1, 'port'=>$2};
   } else {
      return &LoadServerDescription($server);
   }
}


# LoadServerDescription
#  Loads a description for a server from disk. Reloads the description
#  if the $serverconfig given is current.
#

%::Serverconfigs = ();

sub LoadServerDescription {
   my ($server, $serverconfig) = @_;
   my $hash;

   return -1               # sanity check: server name given?
      if !$server;

   $serverconfig = $::Serverconfigs{$server}
      if !defined($serverconfig) and defined($::Serverconfigs{$server});

   my $filename = $::config->{serversdir}.'/'.$server;

   if ( -r $filename ) {   # if file exists and is readable by me

      # If the entry does not exist or modification time of file is newer,
      # reload it.
      #
      if ( (stat(_))[9] != $serverconfig->{mtime} ) {
         if ( $hash = &ConfigFile::Read($filename, \%ServerConfig::Server) ) {
            $serverconfig = $hash;

            # Add a few items to the hash
            #
            $serverconfig->{name}     = $server;
            $serverconfig->{filename} = $filename;
            $serverconfig->{mtime}    = (stat($filename))[9];

            # Enter the entry in the array of servers present
            #
            $::Serverconfigs{$server} = $serverconfig;

            warn "-- Load of server description $filename succeeded"
               if $Z3950::debug;
            return $serverconfig;
         } elsif ( $serverconfig ) {
            warn "-- Load of server description $filename failed, reverting to old entry"
               if $Z3950::debug;
            return $serverconfig;
         } else {
            warn "-- Load of server description $filename failed"
               if $Z3950::debug;
            return undef;
         }
      } else {
         warn '-- Using existing description for '.$server
            if $Z3950::debug;
         return $serverconfig;
      }
   } else {
      warn "-- Description for server $server not available ($filename)"
         if $Z3950::debug;
      return undef;
   }
}

1;

